import { useForceUpdate } from '@lyonph/react-hooks';
import { useDebugValue, useEffect } from 'react';
import { ExternalSubject } from './types';

function useExternalSubjectInternal<T>(
  subject: ExternalSubject<T>,
  suspense = false,
): T {
  const forceUpdate = useForceUpdate();

  // Subscribe to further updates
  useEffect(() => {
    const unsubscribe = subject.subscribe(forceUpdate);
    return unsubscribe;
  }, [subject, forceUpdate]);

  // This effect makes sure that the component
  // is always up-to-date every re-render
  useEffect(() => {
    // If there's no ongoing request, try to request an update.
    if (!subject.getRequest()) {
      subject.requestUpdate();
      const ongoing = subject.getRequest();
      if (ongoing) {
        forceUpdate();
      }
    }
  }); // No dependencies

  // Suspend UI if there's an ongoing request
  const ongoing = subject.getRequest();
  if (ongoing) {
    if (suspense) {
      throw ongoing.promise;
    }
  } else {
    // Check for tearing
    subject.requestUpdate();
    // Suspend if tear is detected
    const current = subject.getRequest();
    if (current) {
      if (suspense) {
        throw current.promise;
      }
    }
  }

  return subject.getCachedValue();
}

export default function useExternalSubject<T>(
  subject: ExternalSubject<T>,
  suspense = false,
): T {
  const value = useExternalSubjectInternal(subject, suspense);
  useDebugValue(value);
  return value;
}
